#Libraries:
import pandas as pd
import numpy as np
import datetime
import calendar
import requests
import os, sys 
import cv2 as cv
from sklearn import preprocessing
from itertools import combinations
from collections import Counter
from itertools import combinations
from collections import Counter

"""
Get_root_path
"""
#
#Body of the function:
#
def get_root_path(n):
    '''
    This function allows us to iterate over folders to add the path of our root folder
    Params:
    - n (int): the number of times we will iterate to reach the desired folder
    '''
    path = os.path.dirname(os.path.abspath(__file__)) 
    for i in range(n):
        path = os.path.dirname(path)
    print(path)
    sys.path.append(path)


#
#Body of the function:
#
def delete_duplicates(df, column_names = None):
    """
    Delete the rows with duplicated values in a dataframe. 
    Arguments: 
    - df : dataframe
    - column_names : names of the columns to take into account
    Returns the dataframe without the duplicated values and prints the number of deleted rows.
    """   
    quant_dup_rows = np.size(df[df.duplicated(subset= column_names)].index)  # Get the number of duplicated rows
    print(f"{quant_dup_rows} rows have been deleted")   # Print the number of duplicated rows

    return df.drop_duplicates(subset= column_names)


"""
Function to deal with missing values in dataframes
"""
#
#Body of the function:
#
def missing_values(df, method = 'drop', column_names = None):
    """
    Function to deal with missing values in dataframes. Three different options: 1 - Drop the rows with missing values, 
    2 - Fill the missing values with zeros, 3 - Fill the missing values with the average of the column values.
    Arguments: 
        - df : the given dataframe
        - method = 'drop' : drop the missing values, 'zero' : fill with zeros, 'avg' : fill with the average
            Default: 'drop' // 
        - column_names --> the columns to take into account
    Return: The resulting dataframe.
    """   
    if method == 'drop':
        df_res = df.dropna(subset = column_names)   # Drop the rows with missing values in the given columns
    elif method == 'zero':
        for name in column_names:
            df[name] = df[name].fillna(0)    # Fill missing values with zeros
        df_res = df
    else:
        for name in column_names:
            df[name] = df[name].fillna(df[name].mean())   # Fill missing values with the average
        df_res = df  

    return df_res

"""
Save the names of columns in a dataframe and export it to csv.
"""
#
#Body of the function:
#
def save_col_names(data):
    """
    Save the names of the columns of a given dataset in a dataframe and export it to csv. 
    Arguments: data --> the given dataframe
    Return: The new dataframe containing the column names.
    """   
    column_names = pd.DataFrame({'Names':data.columns})   # Creates a dataframe with the column names
    column_names.to_csv('column_names.csv', sep = ';')   # Exports the column names to csv

    return column_names


"""
Unix_to_UTC
"""
#
#Body of the function:
#
def Unix_to_UTC(unixdate):
    '''
    This Function transforms data in UNIX time format to UTC time format.
    Arguments: 
    - unixdate : Date in Unix format in interger type
    Unix time is a system for describing a point in time, 
    it is the number of seconds that have elapsed since the Unix epoch,
    The Unix epoch is 00:00:00 UTC on 1 January 1970. 
    IMPORTANT --> UNIX input has to have less than 11 digits
    '''
    
    int(unixdate)
    if unixdate == unixdate>10000000000:
        print('UNIX input has to have less than 11 digits')
    else:
        return datetime.datetime.utcfromtimestamp(unixdate).strftime('%Y-%m-%d:%H:%M')

"""
Split_datetime
"""
#
#Body of the function:
#
def download_csv(url):
    '''
    This Functions takes a URL with a CSV file and downloads it in the same folder the function is being run.
    The only required parameter is the URL itself.
    '''
    req = requests.get(url)
    url_content = req.content

"""
Split_datetime
"""
#
#Body of the function:
#
def split_datetime(df, column, only_date = False):
    
    '''
    This function receives a dataframe with a column with values 'datetime' look like but in 'object' format and transforms it
    into 'datetime' format and divide every component in different columns to operate with the values. In that way, we will
    have days, months, years, hours and minutes in different columns.
    Parameters:
    - df: dataframe
    - column: dataframe's column with the values we want to transform. Tendrá la forma 'df[datetime]'
    - only_date: 'True' if there isn't time and 'False' if there is date and time in the column.
    '''
    column = pd.to_datetime(column)
    time = pd.DataFrame(column)

    time['day'] = column.dt.day
    time['month'] = column.dt.month
    time['year'] = column.dt.year
    time['weekday'] = column.dt.weekday

    if only_date == False:
        time['hour'] = column.dt.hour
        time['minutes'] = column.dt.minute
    
    df = pd.concat([df, time], axis = 1)
    df = df.loc[:,~df.columns.duplicated(keep='last')]
    
    return df

"""
Max_min 
"""
#
#Body of the function:
#
def  max_min (df,column):
    '''
    This Function receives a dataframe and a specific column of said dataframe, 
    it prints the minimun and maximun value in said column.
    Params: 
    - df : dataframe name
    - column : column that you want to iterate on.
    '''
    a = df[column].min()
    print("The minimun value of this column is:", a)
    b = df[column].max()
    print("The maximun value of this column is:", b)

"""
Nulos_to_0
"""
#
#Body of the function:
#
def null_to_0 (df):
    '''
    This Function receives a dataframe and returns the same dataframe with 
    all the NaN values modified to the interger 0.
    Params: 
    - df : dataframe name
    '''
    df = df.fillna(0)
    return df

"""
Nulos_media
"""
#
#Body of the function:
#
def null_to_mean (df,column):
    '''
    This Function receives a dataframe and a specific column of said dataframe,
    and returns the same dataframe with all the NaN values modified 
    to the mean value of that column.
    Params: 
    - df : dataframe name
    - column : column that you want to iterate on.
    '''
    mean_df = df[column].mean()
    df = df.fillna(mean_df)
    return df

"""
Numeric
"""
#
#Body of the function:
#

def categorical_numeric(df, column):
    '''This function is to change the categorical values ​​to numeric values ​​of a concrete column.
    Params:
    - df = DataFrame
    - columns = column to modify
    '''
    le = preprocessing.LabelEncoder()
    for i in column:
        le.fit(df[i])
        df[i] = le.transform(df[i])
    return column

"""
Df_information
"""
#
#Body of the function:
#
def df_information(df):
    '''
    Function to see all the relevant information of the DataFrame
    '''
    print(df.info())
    print(df.describe())
    print(df.corr())
    print(df.head())

"""
Merge_df
"""
#
#Body of the function:
#
def merge_df(df_1, df_2, column):
    '''
    Function to unify 2 DataFrames
    Params:
    - df_n : Each of the Data Frames
    - column : Same column in both dataframes
    '''
    df3 = pd.merge(df_1, df_2, column)
    return df3

"""
Drop_NaN
"""
#
#Body of the function:
#
def drop_NaN(df, column):
    '''This function eliminates the NaN values of a specific column
    Params:
    - df = DataFrame
    - column = columna de la que queremos eliminar los valorers NaN
    '''
    column.replace(0,np.nan)
    df.dropna(axis=0, inplace=True)

"""
Deleate_column
"""
#
#Body of the function:
#
def deleate_column(df,column):
    '''
    Deleting columns and automatically saving to the dataframe
    Params:
    - df:pandas Dataframe
    - col: column to delete
    - axis:1
    - inplace:True
    '''
    df.drop(column,axis=1,inplace=True)

"""
Replace_by_numpy
"""
#
#Body of the function:
#
def replace_by_numpy(df,column,string, value):
    '''
    Replacing string or number with a new value
    Params:
    - df:pandas Dataframe
    - col:column where is the value to replace
    - string: value to replace
    - value:new value
    '''
    df[column] = np.where(df[column] == string, value,df[column])
    return 

"""
Replace_string
"""
#
#Body of the function:
#
def replace_string(df, column, string_list, replacement):
    '''
    Run through list of strings to replace for a value.
    Params:
    - df: pandas Dataframe
    - col: replaced column
    - string_list: list of strings to be runned
    - replacement: new value
    '''
    for pos, val in enumerate(string_list):
        df[column] = df[column].str.replace(string_list[pos], replacement)

"""
Percentage to decimal column
"""
#
#Body of the function:
#
def percentage(df,column):   
    '''
    This functions become your column with % to decimal column of your data.
    Params:
    - df: dataframe
    - column: column to be represented
    '''
    x = df[column].str.replace('%', '').astype(float)
    df[column] = x/100
    return df[column]

"""
Sort_column
"""
#
#Body of the function:
#
def sort_column(df, column):
    '''
    This function sorts a DataFrame according to the selected column. 
    You only have to indicate the columns that you want to make up the dataframe.
    Params:
    - df : Dataframe 
    - column : column you want to modify
    '''
    df = df.sort_values(by=column,ascending=False)
    return df

"""
Outliers_in_column
"""
#
#Body of the function:
#
def outliers (df, column):
    '''
    Function that removes outliers from the given column. 
    Params: 
        - df : Dataframe 
        - column : column you want to modify
    '''
    mean = df[column].mean() 
    std = df[column].std() 
    values = []
    for i in df[column]:
        if i >= (mean + (2*std)):
            i == mean
        values.append(i)
    return values

"""
Get the number of Sales by Category 
"""
#
#Body of the function:
#
def sales_by_category(df, column_1, column_2):
    ''' 
    It will sort values by descending order, from most to least, and top 10. 
    It will create another dataframe that will show column_1, column you want to groupby, 
    column_2 is the number of sales based on column_1 groupby.
    Params: 
    - df : dataframe name
    - column_1 : column that you want to groupby, could be country, state, product 
    - column_2 : column which will indicate a number of sales
    '''
    sales_by_category = df[[column_1,column_2]].groupby([column_1]).count().sort_values(column_2, ascending=False).head(10)
    sales_by_category = sales_by_category.rename(columns={column_2: 'Number of Sales'})
    sales_by_category = sales_by_category.reset_index()
    sales_by_category

    return sales_by_category

"""
From_month_into_to_name
"""
#
#Body of the function:
#
def month_to_name(df, column): 
    '''
    It will return a values of Month in name format for example, 03 -> March 
    Params: 
    - df : dataframe
    - column: should be month, in datatime format
    '''
    df[column] = df[column].apply(lambda x: calendar.month_abbr[x])
    return df

"""
Columns_dtype
"""
#
#Body of the function:
#
def columns_dtype(df, include='all', cardinality_threshold=None, return_df=False):
    '''
    Returns categorical, numerical or both kind of variables based on their 
    dtype, list-like or DataFrame, based on the user specified parameter.
    Params:
        - df: Pandas DataFrame
            Pandas DataFrame where we want to analyze the variables.
        - include: str, default 'all'
            String that indicates which dtype the user wants to return.
            'all' returns both numerical and categorical variables.
            'categorical' returns only categorical variables.
            'numerical' returns only numerical variables.
        - cardinality_threshold: int, default None
            Number of maximum unique values that the categorical variable 
            has to be. The variables with higher cardinality that the 
            specified by this parameter will be ignored.
        - return_df: bool, default False
            Whether the return will be a DataFrame or not. If True, returns 
            the Pandas DataFrame with only the specified variables. If False, 
            returns a list for each of the specified variables.
    Returns:
        - List of variables or DataFrame, depending of the return_df parameter.
    '''

    # Capturing the variables depending of the user's selection

    # Include All
    if include == 'all':
        if cardinality_threshold:
            categorical_cols = [col for col in df.columns if df[col].dtype == 'object' and df[col].nunique() <= cardinality_threshold]
        else:
            categorical_cols = [col for col in df.columns if df[col].dtype == 'object']

        numerical_cols = [col for col in df.columns if df[col].dtype in ['float64', 'int64']]
        data_columns = categorical_cols + numerical_cols

        if return_df:
            return df[data_columns]

        return categorical_cols, numerical_cols

    # Include categorical only
    elif include == 'categorical':
        if cardinality_threshold:
            categorical_cols = [col for col in df.columns if df[col].dtype == 'object' and df[col].nunique() <= cardinality_threshold]
        else:
            categorical_cols = [col for col in df.columns if df[col].dtype == 'object']

        if return_df:
            return df[categorical_cols]

        return categorical_cols

    # Include numerical only
    elif include == 'numerical':
        numerical_cols = [col for col in df.columns if df[col].dtype in ['float64', 'int64']]

        if return_df:
            return df[numerical_cols]

        return numerical_cols

"""
Get_products_that_are_sold_together
"""
#
#Body of the function:
#
def get_products_that_are_sold_together(df, order_id_column, product_column, products_bought_together_column):
    '''
    This function will return a dataframe along with number of transactions of products
    that were purchased together. 
    Params: 
        df : dataframe that will be used 
        order_id_column : is the transaction id, could be order ID. 
        product_column : is the column for products 
        products_bought_together_column : the column that will two products purchased together
    '''
    # creating a new dataframe to separate duplicates from Order ID
    df_updated = df[df[order_id_column].duplicated(keep=False)]
    #Joining the products with the same Order ID group to be on the same line 
    df_updated[products_bought_together_column] = df_updated.groupby(order_id_column)[product_column].transform(lambda x: ','.join(x))
    #Getting rid of the duplicate values 
    df_updated = df_updated[[order_id_column, products_bought_together_column]].drop_duplicates()

    count = Counter()

    for row in df_updated[products_bought_together_column]:
        row_list = row.split(',')
        count.update(Counter(combinations(row_list,2)))

    # Below Will return products bought together along with the number of transactions
    products_bought_together = count.most_common(20)
    products_bought_together_df = pd.DataFrame(products_bought_together, columns=['Product_bundle', 'Number of transactions'])
    return products_bought_together_df


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

# Images

"""
Frames_from_video
"""
#
#Body of the function:
#
def frames_from_video(path_in, path_out, ms_extract = 1000):
    '''
    This function extracts frames from videos in an explicit path, modifies the resolution of the frames and, finally, 
    it saves them in another path.
    This function works well on windows, 
    in MAC it is recommended that you delete the file that is generated automatically (.DS_Store).
    Parameters:
    - path_in: Path of the folder where videos are.
    - path_out: Path of the folder where frames will be saved.
    - ms_extract: miliseconds between frames.
    '''
    filenames = os.listdir(path_in)
    count = 0
    for filename in filenames:
        vidcap = cv.VideoCapture(path_in + '\\' + filename)
        print(path_in + '\\' + filename)
        counter = 0
        while True:
            vidcap.set(cv.CAP_PROP_POS_MSEC,(counter*ms_extract))
            success,image = vidcap.read()
            if success:
                print ('Read a new frame: ', success, count)
                imagesmall = cv.resize(image, (int(image.shape[1]*0.5), int(image.shape[0]*0.5)))                                         
                cv.imwrite( path_out + "\\frame_{}.jpg".format(count), imagesmall)
                count = count + 1
                counter = counter + 1
            else:
                print('***********************')
                print('All frames catched.')
                print('***********************')
                break

    return 'All videos have been catched'



"""
Images_dataset_properties
"""
#
#Body of the function:
#
def images_dataset_properties(images):
    '''
    This function receives an array of images and returns its .ndim, .shape y .size
    This function works well on windows, 
    in MAC it is recommended that you delete the file that is generated automatically (.DS_Store).
    Parameters:
     - images: array with the images. It is an iterable.
    '''
    print("Dimensions:",images.ndim)
    print("Shape:",images.shape)
    print("Size:",images.size)


"""
Read_images
"""
#
#Body of the function:
#
def read_images(path):
    '''
    This function reads an image from a path.
    Parameters:
    - path: Path where the image is.
    '''
    image = cv.imread(path)
    return image

"""
Flip_images
"""
#
#Body of the function:
#
def flip_images(paths_in, paths_out):
    '''
    This function receives the paths where the images are and flip them (mirror effect).
    This function works well on windows, 
    in MAC it is recommended that you delete the file that is generated automatically (.DS_Store).
    Parameters:
    - paths_in: Paths of the folders where the images are. It is an iterable.
    - paths_out: Paths of the folders where the images will be saved. It is an iterable.
    '''

    for i in range(len(paths_in)):
        filenames = os.listdir(paths_in[i])
        count = 0
        for file in filenames:
            if file not in os.listdir(paths_out[i]):
                try: 
                    print('Flipping: ', count+1)
                    path_img = paths_in[i] + '\\' + file
                    image = read_images(path_img)
                    flip = cv.flip(image, 1)
                    cv.imwrite(paths_out[i] + '\\' + file, flip)
                    count = count + 1
                except:
                    break
            else:
                pass

        print('***********************')
        print('Images in folder {} fipped.'.format(paths_in[i]))
        print('***********************')


"""
Resize_images
"""
#
#Body of the function:
#
def resize_images(images, height, width):
    '''
    This function receives images and returns them resized with the specified dimensions.
    This function works well on windows, 
    in MAC it is recommended that you delete the file that is generated automatically (.DS_Store).
    Observacion
    Parameters:
- images: array of images. It is an iterable. If we have just an image, it must be written as 'np.array([image])'
    - height: height in ppp
    - width: width in ppp
    '''
    resized_images = []

    for image in images:
        resized = cv.resize(image, (height, width))
        resized_images.append(resized)
    
    return np.array(resized_images)


"""
Edit_images
"""
#
#Body of the function:
#
def edit_images(images, alpha = 1, beta = 0):
    '''
    This function modifies contrast and bright of an array of images.
    This function works well on windows, 
    in MAC it is recommended that you delete the file that is generated automatically (.DS_Store).
    Parámetros:
    - images: array of images. It is an iterable. If we have just an image, it must be written as 'np.array([image])'
    - alpha: (1.0-3.0). Contrast controller. The higher it is, the higher is the contrast.
    - beta: (0-100). Bright controller. The higher it is, the higher is the bright.
    '''
    edited_images = []

    for image in images:

        adjusted = cv.convertScaleAbs(image, alpha=alpha, beta=beta)
            
        edited_images.append(adjusted)
    
    return np.array(edited_images)

"""
Color2gray
"""
#
#Body of the function:
#
def color2gray(images):
    '''
    Esta función recibe las imágenes en un iterable y las devuelve en escala de grises.
    This function works well on windows, 
    in MAC it is recommended that you delete the file that is generated automatically (.DS_Store).
    Parámetros:
    - images: array of images. It is an iterable. If we have just an image, it must be written as 'np.array([image])'
    '''
    grey_images = []

    for image in images:
        grey = cv.cvtColor(image, cv.COLOR_BGR2GRAY)
        grey_images.append(grey)
    
    return np.array(grey_images)


"""
Negative_colors_images
"""
#
#Body of the function:
#
def negative_colors_images(images): 
    '''
    This function receives an array of images and returns it with the negative colors.
    This function works well on windows, 
    in MAC it is recommended that you delete the file that is generated automatically (.DS_Store).
    Parameters:
    - images: array of images. It is an iterable. If we have just an image, it must be written as 'np.array([image])'
    '''
    negative_images = []

    for image in images:
        negative = 255 - image
        negative_images.append(negative)
    
    return np.array(negative_images)


"""
Monocolor_images
"""
#
#Body of the function:
#
def monocolor_images(images, color = 'blue'):
    '''
    This function receives an array of images and the color channel (red, green o blue) and returns them into this channel.
    Cambia a azul ('blue') si no se especifica.
    This function works well on windows, 
    in MAC it is recommended that you delete the file that is generated automatically (.DS_Store).
    Parameters:
    - images: array of images. It is an iterable. If we have just an image, it must be written as 'np.array([image])'
    - color: color channel we want to get (red, green, blue).
    '''
    monocolor_images = []

    for image in images:

        if color == 'blue':
            b = image.copy()
            b[:,:,0] = b[:,:,1] = 0
            channel = b
        
        elif color == 'green':
            g = image.copy()
            g[:,:,0] = g[:,:,2] = 0
            channel = g
        
        elif color == 'red':
            r = image.copy()
            r[:,:,1] = r[:,:,2] = 0
            channel = r
        
        else:
            print('Error: Chose the correct color.')
            break
        
        monocolor_images.append(channel)

    return np.array(monocolor_images)


"""
Images_load_data
"""
#
#Body of the function:
#
def images_load_data(path_in):
    '''
    This function loads the images from a list of specific locations and assigns them to the variables 'images' and 'labels', so that it returns
    the 'train' and the 'tests' of each of them in different variables. The return would be (X_train, y_train), (X_test, y_test).
    For this to work, the folders must be organized as follows:

        for the train: '../location/train_images' and inside will be the folders with images of each class, divided into those folders, for example: buildings, forest, mountain, beach
        for the test: '../location/test_images' and inside will be the folders with images of each class, divided into those folders, for example: buildings, forest, mountain, beach

    Function call example:
        (train_images, train_labels), (test_images, test_labels) = load_data (pathIn = [r '.... \ images \ train_images', r '.... \ images \ test_images'])

    Variables:
    - path_in: List with all locations. It is an iterable
    '''
    class_names = os.listdir(path_in[0])
    class_names_label = {class_name:i for i, class_name in enumerate(class_names)}

    output = []
    
    # Iterate through training and test sets
    for dataset in path_in:
        
        images = []
        labels = []
        
        print("Loading {}".format(dataset))
        
        # Iterate through each folder corresponding to a category
        for folder in os.listdir(dataset):
            label = class_names_label[folder]
            
            # Iterate through each image in our folder.
            for file in os.listdir(os.path.join(dataset, folder)):
                
                # Get the path name of the image
                img_path = os.path.join(os.path.join(dataset, folder), file)
                
                # Open and resize the img
                image = cv.imread(img_path)
                image = cv.cvtColor(image, cv.COLOR_BGR2RGB)
                image = cv.resize(image, (int(image.shape[1]*0.6), int(image.shape[0]*0.6))) 
                
                # Append the image and its corresponding label to the output
                images.append(image)
                labels.append(label)
                
        try:
            images = np.array(images, 'float32')
            labels = np.array(labels, 'int32')
        
        except:
            images = np.array(images)
            labels = np.array(labels)   
        
        output.append((images, labels))

    return output

"""
Images_load_data
"""
#
#Body of the function:
#
def save_images(image, path_out, image_name):
    '''
    This function saves specified image
    Params:
    - image: image to save
    - path_out: Path of the folder where the image will be saved.
    - image_name: name with the image will be saved
    '''
    cv.imwrite(path_out + "\\" + image_name + ".jpg", image)

    



### LIBRARIES ###
import matplotlib.pyplot as plt
import seaborn as sns 
import pandas as pd
import numpy as np
from scipy import stats
from sklearn.metrics import roc_curve, r2_score, mean_absolute_error, mean_squared_error, accuracy_score, f1_score, precision_score, recall_score,\
                            roc_auc_score, roc_curve
import plotly.express as px
import folium
from IPython.display import display



### FUNCTIONS ###

def draw_boxplot(df, column, color = "b", figsize=(8,8), title = None, label_column_name = None):
    '''
    This function makes a box plot for a specific column.
    ------------------
    Args:
        df: dataframe
        column: column to be represented.
        color: color to be used.
        figsize = here we define the graph size, it has to be a tuple with 2 values
        title: graph name.
        label_column_name: column to be graphed.
    '''
    
    plt.figure(figsize=figsize)
    sns.boxplot(x=df[column], color=color)
    plt.title(title)
    plt.xlabel(label_column_name)
    plt.show()





def sort(df, column):
    '''
    This function sorts a DataFrame according to the selected column. 
    ------------------
    Args:
        df: Pandas DataFrame
        column: here we write the column we want to sort the dataframe by
    ------------------
    Return:
        the same dataframe, only ordered by the selected column
    '''
    df = df.sort_values(by=column,ascending=False)
    return df




def pairplot_heatmap(df):
    
    '''
    Display function that plots a pairplot and a heatmap at the same time showing a pearson correlation at the top.
    ------------------
    Args:
        df = dataframe with numeric variables
    '''

    def codot(*args, **kwargs):
        corr_r = args[0].corr(args[1], 'pearson')
        corr_text = f"{corr_r:2.2f}".replace("0.", ".")
        ax = plt.gca()
        ax.set_axis_off()
        marker_size = abs(corr_r) * 10000
        ax.scatter([.5], [.5], marker_size, [corr_r], alpha=0.6, cmap="coolwarm",
                   vmin=-1, vmax=1, transform=ax.transAxes)
        font_size = abs(corr_r) * 40 + 5
        ax.annotate(corr_text, [.5, .5,],  xycoords="axes fraction",
                    ha='center', va='center', fontsize=font_size)

    g = sns.PairGrid(df, aspect=1.4, diag_sharey=False)
    g.map_lower(sns.regplot, lowess=True, ci=False, line_kws={'color': 'black'})
    g.map_diag(sns.distplot, kde_kws={'color': 'black'})
    g.map_upper(codot);

    return g




def draw_missing_ratio(df, style = 'classic', figsize=(10,5), cmap='inferno', color = "lightgrey", lw=2,  title = None ,fontsize=18):
    ''' 
    This function shows a heatmap with missing values ratio of a dataframe
    ----------------
    Args:
        df = here we define the dataframe we want to analize
        style = here we define the graph style we want to use. 
                To see all the available styles, please type this:  print(plt.style.available)
        figsize = here we define the graph size, it has to be a tuple with 2 values
        cmap = here we define the graph's color palette. To change it, please see the seaborn library
        color = here we define the color of the vertical separation bars
        lw = here we define the thickness of the vertical separation bars
        title = here we define the graph's title, by default is set to None
        fontsize = here we define the title size, if there is a title
    '''
    plt.style.use(style)

    plt.figure(figsize=figsize)

    ax = sns.heatmap(df.isna(),yticklabels=False,xticklabels=df.columns,cbar=False,cmap=cmap)

    ## vertical separation bars between the dataframe columns
    for i in range(df.isna().shape[1]+1):
        ax.axvline(i,color = color, lw=lw);
        
    plt.title(title, fontsize=fontsize);
    plt.show()
    



def draw_statistic_values(df, figsize=(10,8), palette="crest", s= 500, alpha=0.8,  title = None ,fontsize=18, loc_legend= "upper left"):
    ''' 
    This function shows in a scatterplot the 5 most common statistic measures of each numeric column of the dataframe:
        mean: the average
        min= the minimum value
        max= the maximum value
        50%= the median
        std= the standard deviation
    ------------------------
    Args:
        df= the dataframe
        figsize = here we define the graph size, it has to be a tuple with 2 values
        palette= here we define the graph's color palette. To change it, please see the seaborn library
        s= here we define the size of each symbol
        alpha= here we define the transparency of the symbols
        title= here we define th etitle of the graph. It's set to None by default
        fontsize= here we define the title size, if there is a title
        loc_legend= here we define the position of the legend
    ------------------------
    Result:
        a scatterplot with these 5 statistic measures for each column of the dataframe
    '''
    
    fig = plt.figure(figsize=figsize)
    ax = plt.axes()

    ax.set_title(title);
    
    df_stats = df.describe().T
    df_stats = df_stats[["mean","min","max", "50%", "std"]]

    sns.scatterplot(data=df_stats, palette=palette, s= s, alpha=alpha );

    plt.legend(loc=loc_legend);





def draw_target_transformation(column, figsize=(15,5), color = "b"):
            
    ''' 
    This function shows the distribution of a dataframe specific column. 
    Usually, this column is the target (for example in a machine learning problem)
    but, actually, we could apply it to any column with NUMERICAL values.
    
    IMPORTANT: box cox transformation raise an error with 0 values or negative ones.
    
    This function is useful when, in a machine learning problem, we want to see if the distribution of the target column
    is a normal one, or if we have to apply some others transformation to obtain best results.
    ----------------
    Args:
        column = here we define the column we want to see the transformations of. Value must be inserted as Pandas Series
        figsize = here we define the graph size, it has to be a tuple with 2 values
        color = here we define the bars color. Default value is blue ("b")
    ----------------
    Results:
        4 graphs: in the first one we'll see the original distribution, in the second one we'll see the logaritmic distribution,
        in the third one we'll see the boxcox distribution and in the last graph we'll see the power distribution.
    '''
    
    fig,axes = plt.subplots(1, 4, figsize=figsize, sharey=True)

    # Original target
    sns.histplot(column, kde=False, color= color, ax=axes[0])
    axes[0].set_title("Original target")

    # Logaritmic
    sns.histplot(np.log(column),kde=False, color= color, ax=axes[1])
    axes[1].set_title("Log")

    # Box-cox
    #here we define a try/except to manage some critical values
    try:
            sns.histplot(stats.boxcox(column)[0],kde=False,color= color,  ax=axes[2])
            axes[2].set_title("Box-Cox");
    except:
            print("To visualize the boxcox graphs, values must be positive and different from zero.")

    # Power 2
    sns.histplot(np.power(column, 2),kde=False, color= color, ax=axes[3])
    axes[3].set_title("Power 2");





def draw_feat_importance(importance,columns,model_type, figsize=(10,8)):
    '''   
    This function shows a graph with feature importance values of a trained model.
    ------------------
    Args:
        importance: here we have to put the trained model with the feature importance function, we have to write the model, followed by this sentence:  .feature_importances_
            (for example, if we have a trained random forest model, called "rf", the name will be this:  rf.feature_importances_ ).
        columns: here we put the name of the columns we want to show the feature importances of
                (usually, all the dataframe columns, except the target one).
        model_type: here we put the name of model we used to train our dataframe
                (for example random forest, xgboost, etc...)
        figsize: here we define the graph size, it has to be a tuple with 2 values.
    ------------------
    Result:
        the function shows a horizontal bar plot with all the feature importances, sorted by descending order
    
    '''
    
    #Create arrays from feature importance and feature names
    feature_importance = np.array(importance)
    feature_names = np.array(columns)

    #Create a DataFrame using a Dictionary, to store the feature names and their respective feature importances
    data={'feature_names':feature_names,'feature_importance':feature_importance}
    feat_imp_df = pd.DataFrame(data)

    #Sort the DataFrame in order decreasing feature importance
    feat_imp_df.sort_values(by=['feature_importance'], ascending=False,inplace=True)

   
    plt.figure(figsize=figsize)
    sns.barplot(x=feat_imp_df['feature_importance'], y=feat_imp_df['feature_names'])
    plt.title(model_type + ' FEATURE IMPORTANCE')
    plt.xlabel('FEATURE IMPORTANCE')
    plt.ylabel('FEATURE NAMES')






def draw_maps(latitude = 40.4167 , longitude = -3.70325, zoom = 6):
    '''
    This function displays a map based on the latitude, longitude, and zoom.
    By default it will show Spain, with a zoom of 6.
    ------------------
    Args: 
        latitude: here we enter the latitude value, in decimal format
        longitude: here we enter the longitude value, in decimal format
        zoom: here we enter the zoom value of the map
    '''
    
    center = [latitude, longitude]
    my_map = folium.Map(location=center, zoom_start=zoom)
    display(my_map)



  

def draw_sunburst(data_frame, path, color):
    '''
    This function performs a sunburst graph. 
    ------------------
        Args: 
        -"data_frame": the dataframe we want to use
        -"path": we must introduce (as a list) the two columns or variables we want to represent
        -"color": it refers to the column or main variable, that will determine the color (the hue) of the graph
    '''

    fig = px.sunburst(
    data_frame = data_frame,
    path = path,
    color = color,
    color_discrete_sequence = ["red","green","blue","orange"],
    maxdepth = -1,
    )
    fig.update_traces(textinfo='label+percent entry')
    fig.update_layout(margin=dict(t=0, l=0, r=0, b=0))
    fig.show()
    
    



def draw_donut_chart(df, column, title = None):
    ''' This function performs a donut chart. 
    ------------------
        Args: 
            -df = the dataframe 
            -column = enter the column whose two values you want to represent in the chart 
            -title = string format. Here we have to define a title for the graph. By default it's set to None
        '''

    total = df[column].value_counts()
    my_circle=plt.Circle( (0,0), 0.7, color='white') 
    plt.figure(figsize=(10,10))
    plt.pie(total.values,
            labels = total.index,
            autopct='%1.2f%%')
    p = plt.gcf()
    p.gca().add_artist(my_circle)
    plt.title(title)
    plt.show()
    
    



def draw_three_countplot(df, column_ax0, column_ax1, column_ax2, column_hue0, column_hue1, column_hue2, title0, title1,
                     title2, palette1, palette2, palette3):
    '''
    This function performs three vertical countplot graphs with included legend, titles
    and labels with 45 degree rotation and color palette to choose from. 
    ------------------
    Args: 
        -data: the dataframe 
        -column_ax0, column_ax1, column_ax2: here we must enter the columns or variables we want to represent
        -column_hue0, column_hue1, column_hue2: here we must enter the value we want to represent within the coutplot graphs, 
        -title0, title1, title2 : in string format, here we write the titles of each subplot
        -palette1, palette2 and palette3:  these are the color palettes for each subplot
    '''

    fig, axes = plt.subplots(1, 3,  figsize=(20, 8))
    a = sns.countplot(df[column_ax0], hue=df[column_hue0], ax=axes[0], palette= palette1)
    axes[0].set_title(title0)
    a.set_xticklabels(a.get_xticklabels(), rotation=45)
    b = sns.countplot(df[column_ax1], hue=df[column_hue1], palette=palette2, ax=axes[1])
    axes[1].set_title(title1)
    b.set_xticklabels(b.get_xticklabels(), rotation=45)
    c = sns.countplot(df[column_ax2], hue=df[column_hue2], palette=palette3, ax=axes[2])
    axes[2].set_title(title2)
    c.set_xticklabels(c.get_xticklabels(), rotation=45)
    plt.show()




def show_roc_curve(y, y_pred, style = 'seaborn', figsize=(10,5), extra_title = ''):
    ''' 
    This function plots the ROC curve for a classification model predicts 
    ------------------ 
    Args:
        y (pandas.Series): The real target values.
        y_pred (pandas.Series): The target values predicted by the model.
        style (str): Here we define the graph style we want to use. 
            To see all the available styles, please type this: print(plt.style.available)
        figsize (tuple): Here we define the graph size, it has to be a tuple with 2 values
        extra_title (str): An extra text added to the title
    ------------------
    Returns:
        None
    '''

    fpr, tpr, thresholds = roc_curve(y, y_pred, )

    plt.style.use(style)
    plt.figure(figsize=figsize)

    ax = sns.lineplot(fpr, tpr)
    ax.set(xlim = [0.0, 1.0],
           ylim = [0.0, 1.0],
           title = 'ROC curve ' + extra_title,
           xlabel = 'False Positive Rate (1 - Specificity)',
           ylabel = 'True Positive Rate (Sensitivity)',
    )
    plt.show()




# Libraries:
from tensorflow import keras
from sklearn.utils import shuffle
from sklearn.linear_model import LinearRegression, Ridge, Lasso, ElasticNet, LogisticRegression
from sklearn.preprocessing import PolynomialFeatures, LabelEncoder, MinMaxScaler, StandardScaler
from sklearn.pipeline import Pipeline
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier
from sklearn.svm import SVC
from sklearn.model_selection import GridSearchCV, cross_val_score, train_test_split
from sklearn.naive_bayes import GaussianNB
from sklearn.tree import DecisionTreeClassifier
from sklearn.svm import SVC
from sklearn.neighbors import KNeighborsClassifier
from xgboost import XGBClassifier
from sklearn.metrics import accuracy_score,confusion_matrix,roc_auc_score,precision_score,recall_score,f1_score,r2_score,mean_squared_error,mean_absolute_error
from sklearn.preprocessing import OneHotEncoder
from supervised.automl import AutoML
import pandas as pd
import numpy as np
import cv2 as cv
import imageio
import urllib.request

# Functions
def grid_search(X_train, y_train):
    
    '''
    This function performs a grid search with RandomForestClassifier, LogisticRegression and SVC
    to find the best parameters after making train_test_split.
    '''

    pipe = Pipeline(steps=[
        ('classifier', RandomForestClassifier())
    ])

    logistic_params = {
        'classifier': [LogisticRegression()],
        'classifier__penalty': ['l1', 'l2']
    }

    random_forest_params = {
        'classifier': [RandomForestClassifier()],
        'classifier__max_features': [1,2,3]
    }

    svm_param = {
        'classifier': [SVC()],
        'classifier__C': [0.001, 0.1, 0.5, 1, 5, 10, 100],
    }

    search_space = [
        logistic_params,
        random_forest_params,
        svm_param
    ]

    clf = GridSearchCV(estimator = pipe,
                    param_grid = search_space,
                    cv = 10)

    clf.fit(X_train, y_train)
    print(clf.best_estimator_)
    print(clf.best_score_)
    print(clf.best_params_)

#-----------------------------------

def models(X_train, X_test, y_train, y_test):

    '''This function compares and prints results of various ML models after doing train_test_split.'''

    models = []
    models.append(['XGBClassifier',XGBClassifier(learning_rate=0.1,objective='binary:logistic',random_state=0,eval_metric='mlogloss')])
    models.append(['Logistic Regression',LogisticRegression(random_state=0)])
    models.append(['SVC',SVC(random_state=0)])
    models.append(['KNeigbors',KNeighborsClassifier()])
    models.append(['GaussianNB',GaussianNB()])
    models.append(['DecisionTree',DecisionTreeClassifier(random_state=0)])
    models.append(['RandomForest',RandomForestClassifier(random_state=0)])
    models.append(['AdaBoostClassifier',AdaBoostClassifier()])
    lst_1 = []
    for m in range(len(models)):
        lst_2 = []
        model = models[m][1]
        model.fit(X_train,y_train)
        y_pred = model.predict(X_test)
        cm = confusion_matrix(y_test,y_pred)
        accuracies = cross_val_score(estimator= model, X = X_train, y = y_train, cv=10)

    # k-fOLD Validation
        roc = roc_auc_score(y_test,y_pred)
        precision = precision_score(y_test,y_pred)
        recall = recall_score(y_test,y_pred)
        f1 = f1_score(y_test,y_pred)
        print(models[m][0],':')
        print(cm)
        print('Accuracy Score: ',accuracy_score(y_test,y_pred))
        print('')
        print('K-Fold Validation Mean Accuracy: {:.2f} %'.format(accuracies.mean()*100))
        print('')
        print('Standard Deviation: {:.2f} %'.format(accuracies.std()*100))
        print('')
        print('ROC AUC Score: {:.2f} %'.format(roc))
        print('')
        print('Precision: {:.2f} %'.format(precision))
        print('')
        print('Recall: {:.2f} %'.format(recall))
        print('')
        print('F1 Score: {:.2f} %'.format(f1))
        print('-'*40)
        print('')
        lst_2.append(models[m][0])
        lst_2.append(accuracy_score(y_test,y_pred)*100)
        lst_2.append(accuracies.mean()*100)
        lst_2.append(accuracies.std()*100)
        lst_2.append(roc)
        lst_2.append(precision)
        lst_2.append(recall)
        lst_2.append(f1)
        lst_1.append(lst_2)

#-----------------------------------

def NN_conv_model(layers, optimizer = 'adam', loss = 'sparse_categorical_crossentropy', metrics = ['accuracy']):
    
    '''
    This function receives the 'layers', 'optimizer', 'loss' y 'metrics' which are necessaries to create a NN or a CNN.
    Parameters:
    - layers: of the NN or CNN
    - optimizer: optimizer of the model. 'adam' by default
    - loss: loss function. 'sparse_categorical_crossentropy' by default.
    - metrics: metrics for the evaluation. It is an iterable. ['accuracy'] by default.
    
    Example of layers for a CNN:
    
    layers = [keras.layers.Conv2D(first_layer_conv, (3,3), activation=activation, input_shape=image_size),
            keras.layers.MaxPooling2D(pool_size=(2,2)),
            keras.layers.Conv2D(second_layer_conv, (3,3), activation=activation),
            keras.layers.MaxPooling2D(pool_size=(2,2)),
            keras.layers.Flatten(),
            keras.layers.Dense(first_layer_NN, activation=activation),
            keras.layers.Dense(second_layer_NN, activation=activation),
            keras.layers.Dense(len(class_names_label), activation=final_activation)
        ]
    '''

    model = keras.Sequential(layers)

    model.compile(optimizer = optimizer,
                loss = loss,
                metrics = metrics
                )
    
    return model


def model_train(model, X, y, batch_size= 64, epochs= 15, validation_split= 0.1, shuffle_ = False, conv = False):

    '''
    This function receives a model and trains it. Then, it returns the history of the trained model.
    Variables:
    - model: prediction model. It must be a NN or a CNN.
    - X: training values. It is an iterable.
    - y: training labels. It is an iterable.
    - batch_size: batch_size parameter of the NN. 64 by default.
    - epoch: epochs for the training model. 15 by default.
    - validation_split: portion for the validation. 0.1 by default.
    - shuffle_: True for making a 'shuffle' of the values. False by default.
    - conv: False if it is a NN and True if it is a CNN (images). False by default.
    '''

    if shuffle_ == True:
        X, y = shuffle(X, y, random_state=42)
    
    if conv == True:
        X = X/255

    history = model.fit(X,
                    y,
                    batch_size = batch_size,
                    epochs = epochs,
                    validation_split = validation_split
                    )
    return history


def load_model(path):
    
    '''
    The function loads a keras model of the path provided.
    Parameters:
    - path: path of the model.
    '''

    model = keras.models.load_model(path)

    return model


def save_model(model, path, model_name):

    '''
    The function saved the keras model received in the path provided.
    Parameters:
    - model: the trained model
    - path: saving path
    - model_name: name of the model
    '''

    model.save(path + '\\' + model_name + '.h5')

#-----------------------------------

def basic_regression_models(model, X_train, y_train, degree = None, alpha = None, l1_ratio = None):
    
    '''This function receives the type of model and 'X','y' train values to train the basic regression models.
    Parameters:
     - model: the type of model the user wants to train. Values: LinearRegression, PolynomialRegression, Ridge, Lasso, ElasticNet.
     - X_train: X values for training. It has to be an iterable.
     - y_train: yvalues for training. It has to be an iterable.
     - degree: just in case of model = PolynomialRegression. It must be an Integer.
     - alpha: just in case of model = Ridge, Lasso or ElasticNet.
     - l1_ratio: just in case of model = Elasticnet.
     '''
    
    if model == 'LinearRegression':
        lm = LinearRegression()
        trained_model = lm.fit(X_train, y_train)
    
    elif model == 'PolynomialRegression':
        # Preprocessing
        poly_feats = PolynomialFeatures(degree = degree)
        poly_feats.fit(X_train)
        X_poly = poly_feats.transform(X_train)

        # Train
        pol_reg = LinearRegression()
        trained_model = pol_reg.fit(X_poly, y_train)
    
    elif model == 'Ridge':
        ridge = Ridge(alpha = alpha)
        trained_model = ridge.fit(X_train, y_train)
    
    elif model == 'Lasso':
        lasso = Lasso(alpha = alpha)
        trained_model = lasso.fit(X_train, y_train)
    
    elif model == 'ElasticNet':
        elastic = ElasticNet(alpha = alpha, l1_ratio = l1_ratio)
        trained_model = elastic.fit(X_train, y_train)
    
    else:
        print('Error. Chose one of these models: LinearRegression or PolynomialRegression')
    
    return trained_model

#-----------------------------------

def regression_errors(y, y_pred):

    '''
    This function receives the parameters 'y', 'y_pred' and prints the MAE, MSE, RMSE and R2 score.
    Parameters:
    - y: y values for the test
    - y_pred: predicted values from X values for the test
    '''

    print('MAE:', mean_absolute_error(y, y_pred))
    print("MSE:", mean_squared_error(y, y_pred))
    print("RMSE:", np.sqrt(mean_squared_error(y, y_pred)))
    print('R2 score', r2_score(y, y_pred))


def print_regress_metrics(y, y_pred):
    ''' 
    This function print the plot the R^2, MAE, MSE, RMSE and MAPE score of a regression model.
    Args:
        y (pandas.Series): The real target values.
        y_pred (pandas.Series): The target values predicted by the model.
    
    Returns:
        None
    '''

    print("R^2 score:", round(r2_score(y_pred, y), 4))
    print("MAE score:", round(mean_absolute_error(y_pred, y), 4))
    print("MSE score:", round(mean_squared_error(y_pred, y), 4))
    print("RMSE score:", round(np.sqrt(mean_squared_error(y_pred, y)), 4))
    y_array, y_pred_array = np.array(y), np.array(y_pred)
    mape = np.mean(np.abs((y_array - y_pred_array) / y_array)) * 100
    print(f'MAPE score: {round(mape, 4)} %')


def print_classif_metrics(y, y_pred):
    ''' 
    This function print the plot the accuracy, recall, precision, F1 score and AUC
        of a classification model.
    Args:
        y (pandas.Series): The real target values.
        y_pred (pandas.Series): The target values predicted by the model.
    
    Returns:
        None
    '''

    print(f'Accuracy score: {round(accuracy_score(y_pred, y), 3)} %')
    print(f'Recall score: {round(recall_score(y_pred, y), 3)} %')
    print(f'Precision score: {round(precision_score(y_pred, y), 3)} %')
    print(f'F1 score: {round(f1_score(y_pred, y), 3)} %')
    print(f'AUC: {round(roc_auc_score(y_pred, y), 3)} %')

#-----------------------------------

def split_and_scale(data, target_col_name=None, scaling_method='standard', time_series=False, test_size=0.2):

    '''
    Split the data into train and validation data. If the data 
    will train a model for time series, the split will not have 
    shuffling. Once the data is splitted, scale the data (only if 
    the data is not time series data). The user must choose by 
    parameter the scaling method (standard or min-max scaling).
    Params:
        - data: expected Pandas DataFrame.
        - target_col_name: str, default None
            name of the target variable.
        -  scaling_method: str, default 'standard'
            allowed 'minmax' or 'standard'.
        - time_series: bool, default False
            if True, shuffle and scaling will be False.
        - test_size: float, default 0.2
            percentage size of the validation data (0 to 1).
    Returns:
        - Splitted —and scaled, if selected— data. 
    
    '''

    # Getting Features and Target
    X = data.drop(target_col_name, axis=1)
    y = data[target_col_name]

    if time_series:
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=test_size, shuffle=False)
        
    else:
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=test_size, shuffle=True)

        # Scaling the data
        if scaling_method == 'standard':
            scaler = StandardScaler()

            # Scaling the data (Standard Scaler)
            X_train = scaler.fit_transform(X_train)
            X_test = scaler.transform(X_test)

        elif scaling_method == 'minmax':
            scaler = MinMaxScaler()

            # Scaling the data (MinMax Scaler)
            X_train = scaler.fit_transform(X_train)
            X_test = scaler.transform(X_test)

    return X_train, X_test, y_train, y_test

#-----------------------------------

def categorical_encoding(data, kind='label_encoding'):

    '''
    Function to encode categorical variables.
    Encodes categorical data from a given DataFrame with one of the 
    following methods: Label Encoding or One Hot Encoding. This function 
    makes use of the scikit-learn library.
    Params:
        - data: expected Pandas DataFrame
        - kind: str, default 'label_encoding'
            one of both 'label_encoding' or 'one_hot_encoding' must be passed
    Returns:
        - Pandas DataFrame with categorical data encoded
    '''

    # Checking if kind variable is correct:
    if kind not in ['label_encoding', 'one_hot_encoding']:
        raise ValueError ('"kind" parameter must be "label_encoding" or "one_hot_encoding"')

    # Checking the encoding
    if kind == 'label_encoding':
        le = LabelEncoder()
        categorical_columns = [col for col in data.columns if data[col].dtype == 'object']
        for column in categorical_columns:
            data[column] = le.fit_transform(data[column])
        
        # Returning label-encoded data
        return data

    elif kind == 'one_hot_encoding':

        # Getting categorical columns
        categorical_columns = [col for col in data.columns if data[col].dtype == 'object']

        oh = OneHotEncoder(handle_unknown='ignore', sparse=False)

        oh_encoded_dataframe = pd.DataFrame(oh.fit_transform(data[categorical_columns]))
        # Putting back the indexes
        oh_encoded_dataframe.index = data.index
        # Removing categorical columns to replace with onehot-encoded columns
        numeric_data = data.drop(categorical_columns, axis=1)
        # Adding categorical columns to numerical features
        onehot_encoded_data = pd.concat([numeric_data, oh_encoded_dataframe], axis=1)

        # Returning onehot-encoded data
        return onehot_encoded_data

#-----------------------------------

def automl(X_train,y_train,X_test,y_test):

    '''
    This function receives the train and test data and makes an AutoML to compare different prediction model
    with some default values
    '''

    my_automl = AutoML(eval_metric='accuracy')
    my_automl.fit(X_train,y_train)
    preds = my_automl.predict(X_test)
    accuracy_score(preds, y_test)
    return accuracy_score

#-----------------------------------

def captain():
    
    '''
    For calling best function ever you must have installed the libraries 'imageio' and 'urllib'.
    Press 'q' to exit from our Captain.
    '''

    print("Fuck Blackbeard... I'm the captain.")

    url = "https://64.media.tumblr.com/bba725c6dd9e1fdcacc651925c44d0d5/b6dc49731ce9499b-60/s640x960/1192587ef5ad3412f4020077a0c5ae51a387392d.gifv"
    fname = "CAPTAIN Jack Sparrow, please.gif"

    ## Read the gif from the web, save to the disk
    imdata = urllib.request.urlopen(url).read()
    imbytes = bytearray(imdata)
    open(fname,"wb+").write(imdata)

    ## Read the gif from disk to `RGB`s using `imageio.miread` 
    gif = imageio.mimread(fname)
    nums = len(gif)
    print("Total {} frames in the gif!".format(nums))

    # convert form RGB to BGR 
    imgs = [cv.cvtColor(img, cv.COLOR_RGB2BGR) for img in gif]

    ## Display the gif
    i = 0

    while True:
        cv.imshow("gif", imgs[i])
        if cv.waitKey(100)&0xFF==ord('q'):
            break
        i = (i+1)%nums
    cv.destroyAllWindows()


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
  name = 'blackbeard2109',         
  version = '0.2.3',    
  author = 'The Bridge Data Science Team 2109',                   
  author_email = 'datascience2109thebridge@gmail.com',
  description = '''This library is designed for people who need to optimize time in an agile way with an ease of understanding and could 
  solve the main projects you may have in Data Science, starting with cleaning dataframe (including images), visualization and machine learning.''',   
  long_description=long_description,
  long_description_content_type="text/markdown",
  url = 'https://github.com/ds2109fulltime/BLACKBEARD',   
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
  ],
  package_dir={"": "src"},
  packages=setuptools.find_packages(where="src"),
  install_requires=[
    'pandas',
    'numpy',
    'opencv-python',
    'scikit-learn',
    'imageio',
    'folium',
    'plotly',
    'scipy',
    'IPython',
    'matplotlib',
    'seaborn',
    'tensorflow',
    'xgboost',
    'mljar-supervised'
  ],
  python_requires=">=3.6",
)
